from sqlframe.base.function_alternatives import (  # noqa
    any_value_ignore_nulls_not_supported as any_value,
    e_literal as e,
    expm1_from_exp as expm1,
    log1p_from_log as log1p,
    rint_from_round as rint,
    bitwise_not_from_bitnot as bitwise_not,
    skewness_from_skew as skewness,
    isnan_using_equal as isnan,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    percentile_approx_without_accuracy_and_max_array as percentile_approx,
    bround_using_half_even as bround,
    shiftleft_from_bitshiftleft as shiftleft,
    shiftright_from_bitshiftright as shiftright,
    struct_with_eq as struct,
    make_date_date_from_parts as make_date,
    date_add_no_date_sub as date_add,
    date_add_no_date_sub as dateadd,
    date_sub_by_date_add as date_sub,
    add_months_using_func as add_months,
    months_between_cast_as_date_cast_roundoff as months_between,
    last_day_with_cast as last_day,
    from_unixtime_from_timestamp as from_unixtime,
    unix_timestamp_from_extract as unix_timestamp,
    base64_from_base64_encode as base64,
    unbase64_from_base64_decode_string as unbase64,
    format_number_from_to_char as format_number,
    overlay_from_substr as overlay,
    levenshtein_edit_distance as levenshtein,
    split_with_split as split,
    regexp_extract_coalesce_empty_str as regexp_extract,
    hex_using_encode as hex,
    unhex_hex_decode_str as unhex,
    create_map_with_cast as create_map,
    array_contains_cast_variant as array_contains,
    arrays_overlap_as_plural as arrays_overlap,
    slice_as_array_slice as slice,
    array_join_no_null_replacement as array_join,
    array_position_cast_variant_and_flip as array_position,
    element_at_using_brackets as element_at,
    array_intersect_using_intersection as array_intersect,
    array_union_using_array_concat as array_union,
    sort_array_using_array_sort as sort_array,
    flatten_using_array_flatten as flatten,
    map_concat_using_map_cat as map_concat,
    sequence_from_array_generate_range as sequence,
    to_number_using_to_double as to_number,
)
from sqlframe.base.functions import (
    abs as abs,
    acos as acos,
    acosh as acosh,
    approxCountDistinct as approxCountDistinct,
    approx_count_distinct as approx_count_distinct,
    array as array,
    array_distinct as array_distinct,
    array_except as array_except,
    array_max as array_max,
    array_min as array_min,
    array_remove as array_remove,
    array_sort as array_sort,
    asc as asc,
    asc_nulls_first as asc_nulls_first,
    asc_nulls_last as asc_nulls_last,
    ascii as ascii,
    asin as asin,
    asinh as asinh,
    atan as atan,
    atan2 as atan2,
    atanh as atanh,
    avg as avg,
    bit_length as bit_length,
    bitwiseNOT as bitwiseNOT,
    bool_and as bool_and,
    bool_or as bool_or,
    call_function as call_function,
    cbrt as cbrt,
    ceil as ceil,
    ceiling as ceiling,
    char as char,
    coalesce as coalesce,
    col as col,
    collect_list as collect_list,
    collect_set as collect_set,
    concat as concat,
    concat_ws as concat_ws,
    corr as corr,
    cos as cos,
    cosh as cosh,
    cot as cot,
    count as count,
    countDistinct as countDistinct,
    count_distinct as count_distinct,
    count_if as count_if,
    covar_pop as covar_pop,
    covar_samp as covar_samp,
    cume_dist as cume_dist,
    current_date as current_date,
    current_timestamp as current_timestamp,
    current_user as current_user,
    date_diff as date_diff,
    date_format as date_format,
    date_trunc as date_trunc,
    datediff as datediff,
    dayofmonth as dayofmonth,
    dayofweek as dayofweek,
    dayofyear as dayofyear,
    degrees as degrees,
    dense_rank as dense_rank,
    desc as desc,
    desc_nulls_first as desc_nulls_first,
    desc_nulls_last as desc_nulls_last,
    exp as exp,
    explode as explode,
    expr as expr,
    extract as extract,
    factorial as factorial,
    floor as floor,
    greatest as greatest,
    grouping_id as grouping_id,
    hash as hash,
    hour as hour,
    ifnull as ifnull,
    initcap as initcap,
    input_file_name as input_file_name,
    instr as instr,
    kurtosis as kurtosis,
    lag as lag,
    lcase as lcase,
    lead as lead,
    least as least,
    left as left,
    length as length,
    lit as lit,
    ln as ln,
    locate as locate,
    log as log,
    log10 as log10,
    log2 as log2,
    lower as lower,
    lpad as lpad,
    ltrim as ltrim,
    map_keys as map_keys,
    max as max,
    max_by as max_by,
    md5 as md5,
    mean as mean,
    min as min,
    min_by as min_by,
    minute as minute,
    month as month,
    next_day as next_day,
    now as now,
    nth_value as nth_value,
    ntile as ntile,
    nullif as nullif,
    nvl as nvl,
    nvl2 as nvl2,
    octet_length as octet_length,
    percent_rank as percent_rank,
    percentile as percentile,
    posexplode as posexplode,
    position as position,
    pow as pow,
    power as power,
    quarter as quarter,
    radians as radians,
    rand as rand,
    rank as rank,
    regexp_replace as regexp_replace,
    repeat as repeat,
    right as right,
    round as round,
    row_number as row_number,
    rpad as rpad,
    rtrim as rtrim,
    second as second,
    sha as sha,
    sha1 as sha1,
    sha2 as sha2,
    shiftLeft as shiftLeft,
    shiftRight as shiftRight,
    sign as sign,
    signum as signum,
    sin as sin,
    sinh as sinh,
    size as size,
    soundex as soundex,
    sqrt as sqrt,
    startswith as startswith,
    stddev as stddev,
    stddev_pop as stddev_pop,
    stddev_samp as stddev_samp,
    substring as substring,
    sum as sum,
    sumDistinct as sumDistinct,
    sum_distinct as sum_distinct,
    tan as tan,
    tanh as tanh,
    timestamp_seconds as timestamp_seconds,
    toDegrees as toDegrees,
    toRadians as toRadians,
    to_date as to_date,
    to_timestamp as to_timestamp,
    translate as translate,
    trim as trim,
    trunc as trunc,
    ucase as ucase,
    upper as upper,
    user as user,
    var_pop as var_pop,
    var_samp as var_samp,
    variance as variance,
    weekofyear as weekofyear,
    when as when,
    year as year,
)



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlglot import expressions as exp
from sqlglot.helper import ensure_list

from sqlframe.base.column import Column
from sqlframe.base.util import get_tables_from_expression_with_join

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import SESSION

    NORMALIZE_INPUT = t.TypeVar("NORMALIZE_INPUT", bound=t.Union[str, exp.Expression, Column])


def normalize(session: SESSION, expression_context: exp.Select, expr: t.List[NORMALIZE_INPUT]):
    expr = ensure_list(expr)
    expressions = _ensure_expressions(expr)
    for expression in expressions:
        identifiers = expression.find_all(exp.Identifier)
        for identifier in identifiers:
            identifier.transform(session.input_dialect.normalize_identifier)
            replace_alias_name_with_cte_name(session, expression_context, identifier)
            replace_branch_and_sequence_ids_with_cte_name(session, expression_context, identifier)


def replace_alias_name_with_cte_name(
    session: SESSION, expression_context: exp.Select, id: exp.Identifier
):
    normalized_id = session._normalize_string(id.alias_or_name)
    if normalized_id in session.name_to_sequence_id_mapping:
        for cte in reversed(expression_context.ctes):
            if cte.args["sequence_id"] in session.name_to_sequence_id_mapping[normalized_id]:
                _set_alias_name(id, cte.alias_or_name)
                break


def replace_branch_and_sequence_ids_with_cte_name(
    session: SESSION, expression_context: exp.Select, id: exp.Identifier
):
    normalized_id = session._normalize_string(id.alias_or_name)
    if normalized_id in session.known_ids:
        # Check if we have a join and if both the tables in that join share a common branch id
        # If so we need to have this reference the left table by default unless the id is a sequence
        # id then it keeps that reference. This handles the weird edge case in spark that shouldn't
        # be common in practice
        if expression_context.args.get("joins") and normalized_id in session.known_branch_ids:
            join_table_aliases = [
                x.alias_or_name for x in get_tables_from_expression_with_join(expression_context)
            ]
            ctes_in_join = [
                cte for cte in expression_context.ctes if cte.alias_or_name in join_table_aliases
            ]
            if ctes_in_join[0].args["branch_id"] == ctes_in_join[1].args["branch_id"]:
                assert len(ctes_in_join) == 2
                _set_alias_name(id, ctes_in_join[0].alias_or_name)
                return

        for cte in reversed(expression_context.ctes):
            if normalized_id in (cte.args["branch_id"], cte.args["sequence_id"]):
                _set_alias_name(id, cte.alias_or_name)
                return


def normalize_dict(session: SESSION, data: t.Dict) -> t.Dict:
    if isinstance(data, dict):
        return {session._normalize_string(k): normalize_dict(session, v) for k, v in data.items()}
    elif isinstance(data, list):
        return [normalize_dict(session, v) for v in data]
    else:
        return data


def _set_alias_name(id: exp.Identifier, name: str):
    id.set("this", name)
    id.set("quoted", False)


def _ensure_expressions(values: t.List[NORMALIZE_INPUT]) -> t.List[exp.Expression]:
    results = []
    for value in values:
        if isinstance(value, str):
            results.append(Column.ensure_col(value).expression)
        elif isinstance(value, Column):
            results.append(value.expression)
        elif isinstance(value, exp.Expression):
            results.append(value)
        else:
            raise ValueError(f"Got an invalid type to normalize: {type(value)}")
    return results



from __future__ import annotations

import pathlib
import typing as t

from sqlframe.base.exceptions import UnsupportedOperationError
from sqlframe.base.readerwriter import (
    DF,
    SESSION,
    _BaseDataFrameReader,
    _BaseDataFrameWriter,
    _infer_format,
)
from sqlframe.base.util import pandas_to_spark_schema, verify_pandas_installed

if t.TYPE_CHECKING:
    from sqlframe.base._typing import OptionalPrimitiveType, PathOrPaths
    from sqlframe.base.types import StructType


class PandasLoaderMixin(_BaseDataFrameReader, t.Generic[SESSION, DF]):
    def load(
        self,
        path: t.Optional[PathOrPaths] = None,
        format: t.Optional[str] = None,
        schema: t.Optional[t.Union[StructType, str]] = None,
        **options: OptionalPrimitiveType,
    ) -> DF:
        """Loads data from a data source and returns it as a :class:`DataFrame`.

        .. versionadded:: 1.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        path : str or list, t.Optional
            t.Optional string or a list of string for file-system backed data sources.
        format : str, t.Optional
            t.Optional string for format of the data source. Default to 'parquet'.
        schema : :class:`pyspark.sql.types.StructType` or str, t.Optional
            t.Optional :class:`pyspark.sql.types.StructType` for the input schema
            or a DDL-formatted string (For example ``col0 INT, col1 DOUBLE``).
        **options : dict
            all other string options

        Examples
        --------
        Load a CSV file with format, schema and options specified.

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as d:
        ...     # Write a DataFrame into a CSV file with a header
        ...     df = spark.createDataFrame([{"age": 100, "name": "Hyukjin Kwon"}])
        ...     df.write.option("header", True).mode("overwrite").format("csv").save(d)
        ...
        ...     # Read the CSV file as a DataFrame with 'nullValue' option set to 'Hyukjin Kwon',
        ...     # and 'header' option set to `True`.
        ...     df = spark.read.load(
        ...         d, schema=df.schema, format="csv", nullValue="Hyukjin Kwon", header=True)
        ...     df.printSchema()
        ...     df.show()
        root
         |-- age: long (nullable = true)
         |-- name: string (nullable = true)
        +---+----+
        |age|name|
        +---+----+
        |100|NULL|
        +---+----+
        """
        verify_pandas_installed()
        import pandas as pd

        assert path is not None, "path is required"
        assert isinstance(path, str), "path must be a string"
        format = format or _infer_format(path)
        kwargs = {k: v for k, v in options.items() if v is not None}
        if format == "json":
            df = pd.read_json(path, lines=True, **kwargs)  # type: ignore
        elif format == "parquet":
            df = pd.read_parquet(path, **kwargs)  # type: ignore
        elif format == "csv":
            df = pd.read_csv(path, **kwargs)  # type: ignore
        else:
            raise UnsupportedOperationError(f"Unsupported format: {format}")
        schema = schema or pandas_to_spark_schema(df)
        self.session._last_loaded_file = path
        return self._session.createDataFrame(list(df.itertuples(index=False)), schema)


class PandasWriterMixin(_BaseDataFrameWriter, t.Generic[SESSION, DF]):
    def _write(self, path: str, mode: t.Optional[str], format: str, **options):  # type: ignore
        mode, skip = self._validate_mode(path, mode)
        if skip:
            return
        pandas_df = self._df.toPandas()
        mode = self._mode_to_pandas_mode(mode)
        kwargs = {k: v for k, v in options.items() if v is not None}
        kwargs["index"] = False
        if format == "csv":
            kwargs["mode"] = mode
            if mode == "a" and pathlib.Path(path).exists():
                kwargs["header"] = False
            pandas_df.to_csv(path, **kwargs)
        elif format == "parquet":
            if mode == "a":
                raise NotImplementedError("Append mode is not supported for parquet.")
            pandas_df.to_parquet(path, **kwargs)
        elif format == "json":
            # Pandas versions are inconsistent on how to handle True/False index so we just remove it
            # since in all versions it will not result in an index column in the output.
            del kwargs["index"]
            kwargs["mode"] = mode
            kwargs["orient"] = "records"
            pandas_df.to_json(path, lines=True, **kwargs)
        else:
            raise NotImplementedError(f"Unsupported format: {format}")



from __future__ import annotations

import typing as t

from sqlframe.base.udf import _BaseUDFRegistration

if t.TYPE_CHECKING:
    from sqlframe.redshift.session import RedshiftSession


class RedshiftUDFRegistration(_BaseUDFRegistration["RedshiftSession"]): ...



from sqlframe.standalone.catalog import StandaloneCatalog
from sqlframe.standalone.column import Column
from sqlframe.standalone.dataframe import (
    StandaloneDataFrame,
    StandaloneDataFrameNaFunctions,
)
from sqlframe.standalone.group import StandaloneGroupedData
from sqlframe.standalone.readwriter import (
    StandaloneDataFrameReader,
    StandaloneDataFrameWriter,
)
from sqlframe.standalone.session import StandaloneSession
from sqlframe.standalone.window import Window, WindowSpec

__all__ = [
    "StandaloneCatalog",
    "Column",
    "StandaloneDataFrame",
    "StandaloneDataFrameNaFunctions",
    "StandaloneGroupedData",
    "StandaloneDataFrameReader",
    "StandaloneDataFrameWriter",
    "StandaloneSession",
    "Window",
    "WindowSpec",
]



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import fnmatch
import typing as t

from sqlglot import exp, parse_one

from sqlframe.base.catalog import Column, Function, _BaseCatalog
from sqlframe.base.mixins.catalog_mixins import (
    GetCurrentCatalogFromFunctionMixin,
    GetCurrentDatabaseFromFunctionMixin,
    ListCatalogsFromInfoSchemaMixin,
    ListDatabasesFromInfoSchemaMixin,
    ListTablesFromInfoSchemaMixin,
    SetCurrentDatabaseFromSearchPathMixin,
)
from sqlframe.base.util import normalize_string, to_schema

if t.TYPE_CHECKING:
    from sqlframe.postgres.session import PostgresSession  # noqa
    from sqlframe.postgres.dataframe import PostgresDataFrame  # noqa


class PostgresCatalog(
    GetCurrentCatalogFromFunctionMixin["PostgresSession", "PostgresDataFrame"],
    GetCurrentDatabaseFromFunctionMixin["PostgresSession", "PostgresDataFrame"],
    ListDatabasesFromInfoSchemaMixin["PostgresSession", "PostgresDataFrame"],
    ListCatalogsFromInfoSchemaMixin["PostgresSession", "PostgresDataFrame"],
    SetCurrentDatabaseFromSearchPathMixin["PostgresSession", "PostgresDataFrame"],
    ListTablesFromInfoSchemaMixin["PostgresSession", "PostgresDataFrame"],
    _BaseCatalog["PostgresSession", "PostgresDataFrame"],
):
    CURRENT_CATALOG_EXPRESSION: exp.Expression = exp.column("current_catalog")
    TEMP_SCHEMA_FILTER = exp.column("table_schema").like("pg_temp_%")

    def listColumns(
        self, tableName: str, dbName: t.Optional[str] = None, include_temp: bool = False
    ) -> t.List[Column]:
        """Returns a t.List of columns for the given table/view in the specified database.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        tableName : str
            name of the table to t.List columns.

            .. versionchanged:: 3.4.0
               Allow ``tableName`` to be qualified with catalog name when ``dbName`` is None.

        dbName : str, t.Optional
            name of the database to find the table to t.List columns.

        Returns
        -------
        t.List
            A t.List of :class:`Column`.

        Notes
        -----
        The order of arguments here is different from that of its JVM counterpart
        because Python does not support method overloading.

        If no database is specified, the current database and catalog
        are used. This API includes all temporary views.

        Examples
        --------
        >>> _ = spark.sql("DROP TABLE IF EXISTS tbl1")
        >>> _ = spark.sql("CREATE TABLE tblA (name STRING, age INT) USING parquet")
        >>> spark.catalog.t.listColumns("tblA")
        [Column(name='name', description=None, dataType='string', nullable=True, ...
        >>> _ = spark.sql("DROP TABLE tblA")
        """
        tableName = normalize_string(tableName, from_dialect="input", is_table=True)
        if df := self.session.temp_views.get(tableName):
            return [
                Column(
                    name=x,
                    description=None,
                    dataType="",
                    nullable=True,
                    isPartition=False,
                    isBucket=False,
                )
                for x in df.columns
            ]

        table = exp.to_table(tableName, dialect=self.session.input_dialect)
        dbName = normalize_string(dbName, from_dialect="input", is_schema=True) if dbName else None
        schema = to_schema(dbName, dialect=self.session.input_dialect) if dbName else None
        if not table.db:
            if schema and schema.db:
                table.set("db", schema.args["db"])
            else:
                current_database = normalize_string(
                    self.currentDatabase(), from_dialect="output", to_dialect="input"
                )
                table.set(
                    "db",
                    exp.parse_identifier(current_database, dialect=self.session.input_dialect),
                )
        if not table.catalog:
            if schema and schema.catalog:
                table.set("catalog", schema.args["catalog"])
            else:
                current_catalog = normalize_string(
                    self.currentCatalog(), from_dialect="output", to_dialect="input"
                )
                table.set(
                    "catalog",
                    exp.parse_identifier(current_catalog, dialect=self.session.input_dialect),
                )
        select = parse_one(
            f"""
        SELECT
    att.attname AS column_name,
    pg_catalog.format_type(att.atttypid, NULL) AS data_type,
    col.is_nullable
FROM
    pg_catalog.pg_attribute att
JOIN
    pg_catalog.pg_class cls ON cls.oid = att.attrelid
JOIN
    pg_catalog.pg_namespace nsp ON nsp.oid = cls.relnamespace
JOIN
    information_schema.columns col ON col.table_schema = nsp.nspname AND col.table_name = cls.relname AND col.column_name = att.attname
WHERE
    cls.relname = '{table.name}' AND
    att.attnum > 0 AND
    NOT att.attisdropped
ORDER BY
    att.attnum;
        """,
            dialect="postgres",
        )
        if table.db:
            schema_filter: exp.Expression = exp.column("table_schema").eq(table.db)
            if include_temp and self.TEMP_SCHEMA_FILTER:
                schema_filter = exp.Or(this=schema_filter, expression=self.TEMP_SCHEMA_FILTER)
            select = select.where(schema_filter)  # type: ignore
        if table.catalog:
            catalog_filter: exp.Expression = exp.column("table_catalog").eq(table.catalog)
            if include_temp and self.TEMP_CATALOG_FILTER:
                catalog_filter = exp.Or(this=catalog_filter, expression=self.TEMP_CATALOG_FILTER)
            select = select.where(catalog_filter)  # type: ignore
        results = self.session._collect(select)
        return [
            Column(
                name=normalize_string(
                    x["column_name"], from_dialect="execution", to_dialect="output"
                ),
                description=None,
                dataType=normalize_string(
                    x["data_type"], from_dialect="execution", to_dialect="output", is_datatype=True
                ),
                nullable=x["is_nullable"] == "YES",
                isPartition=False,
                isBucket=False,
            )
            for x in results
        ]

    def listFunctions(
        self, dbName: t.Optional[str] = None, pattern: t.Optional[str] = None
    ) -> t.List[Function]:
        """
        Returns a t.List of functions registered in the specified database.

        .. versionadded:: 3.4.0

        Parameters
        ----------
        dbName : str
            name of the database to t.List the functions.
            ``dbName`` can be qualified with catalog name.
        pattern : str
            The pattern that the function name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Function`.

        Notes
        -----
        If no database is specified, the current database and catalog
        are used. This API includes all temporary functions.

        Examples
        --------
        >>> spark.catalog.t.listFunctions()
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="to_*")
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="*not_existing_func*")
        []
        """
        # SO: https://stackoverflow.com/questions/44143816/any-way-to-list-all-user-defined-postgresql-functions
        query = parse_one(
            """SELECT n.nspname as "namespace",
  p.proname as "name"
FROM pg_catalog.pg_proc p
     LEFT JOIN pg_catalog.pg_namespace n ON n.oid = p.pronamespace
WHERE pg_catalog.pg_function_is_visible(p.oid)
      AND n.nspname <> 'pg_catalog'
      AND n.nspname <> 'information_schema'
ORDER BY 1, 2;
        """,
            dialect="postgres",
        )
        functions = self.session._collect(query)
        catalog = normalize_string(self.currentCatalog(), from_dialect="output")
        results = [
            Function(
                name=normalize_string(x["name"], from_dialect="execution", to_dialect="output"),
                catalog=catalog,
                namespace=[
                    normalize_string(x["namespace"], from_dialect="execution", to_dialect="output")
                ],
                description=None,
                className="",
                isTemporary=False,
            )
            for x in functions
        ]
        if pattern:
            normalized_pattern = normalize_string(
                pattern, from_dialect="input", to_dialect="output", is_pattern=True
            )
            results = [x for x in results if fnmatch.fnmatch(x.name, normalized_pattern)]
        return results



import typing as t

import pytest
from pyspark.sql import DataFrame as PySparkDataFrame
from pyspark.sql import SparkSession as PySparkSession
from pyspark.sql import functions as F

pytest_plugins = ["tests.integration.fixtures"]


def test_sql_simple_select(
    pyspark_session: PySparkSession,
    pyspark_employee: PySparkDataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
):
    employee = get_df("employee")
    query = "SELECT fname, lname FROM employee"
    df = pyspark_session.sql(query)
    dfs = employee.session.sql(query)
    compare_frames(df, dfs)


def test_sql_with_join(
    pyspark_session: PySparkSession,
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    get_func: t.Callable,
):
    employee = get_df("employee")
    col = get_func("col", employee.session)
    countDistinct = get_func("countDistinct", employee.session)
    store = get_df("store")
    query = """
    SELECT
        e.employee_id
        , s.store_id    
    FROM
        employee e
        INNER JOIN
        store s
        ON
            e.store_id = s.store_id
    """
    df = (
        pyspark_session.sql(query)
        .groupBy(F.col("store_id"))
        .agg(F.countDistinct(F.col("employee_id")))
    )
    dfs = (
        employee.session.sql(query).groupBy(col("store_id")).agg(countDistinct(col("employee_id")))
    )
    compare_frames(df, dfs, compare_schema=False)


def test_nameless_column(
    pyspark_session: PySparkSession,
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
):
    employee = get_df("employee")
    query = "SELECT MAX(age) FROM employee"
    df = pyspark_session.sql(query)
    dfs = employee.session.sql(query)
    # Spark will alias the column to `max(age)` while sqlglot will alias to `_col_0` so their schemas will differ
    compare_frames(df, dfs, compare_schema=False)



from __future__ import annotations

import typing as t

from sqlglot.helper import ensure_list

CALLING_CLASS = t.TypeVar("CALLING_CLASS")


def func_metadata(unsupported_engines: t.Optional[t.Union[str, t.List[str]]] = None) -> t.Callable:
    def _metadata(func: t.Callable) -> t.Callable:
        func.unsupported_engines = ensure_list(unsupported_engines) if unsupported_engines else []  # type: ignore
        return func

    return _metadata



from sqlframe.duckdb import DuckDBSession

pytest_plugins = ["tests.integration.fixtures"]


def test_udf_register(duckdb_session: DuckDBSession):
    def multi(x: int, y: int) -> int:
        return x * y

    duckdb_session.udf.register("multi", multi)

    assert duckdb_session.sql("SELECT multi(1, 2)").collect() == [(2,)]



import typing as t

import pytest
from pyspark.sql import DataFrame
from pyspark.sql import functions as F

pytest_plugins = ["tests.integration.fixtures"]


def test_group_by(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    get_func: t.Callable,
):
    employee = get_df("employee")
    min = get_func("min", employee.session)
    df_employee = pyspark_employee.groupBy(pyspark_employee.age).agg(
        F.min(pyspark_employee.employee_id)
    )
    dfs_employee = employee.groupBy(employee.age).agg(min(employee.employee_id))
    compare_frames(df_employee, dfs_employee, compare_schema=False)


def test_group_by_where_non_aggregate(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    get_func: t.Callable,
):
    employee = get_df("employee")
    min = get_func("min", employee.session)
    col = get_func("col", employee.session)
    lit = get_func("lit", employee.session)
    df_employee = (
        pyspark_employee.groupBy(pyspark_employee.age)
        .agg(F.min(pyspark_employee.employee_id).alias("min_employee_id"))
        .where(F.col("age") > F.lit(50))
    )
    dfs_employee = (
        employee.groupBy(employee.age)
        .agg(min(employee.employee_id).alias("min_employee_id"))
        .where(col("age") > lit(50))
    )
    compare_frames(df_employee, dfs_employee)


def test_group_by_where_aggregate_like_having(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    get_func: t.Callable,
):
    employee = get_df("employee")
    min = get_func("min", employee.session)
    col = get_func("col", employee.session)
    lit = get_func("lit", employee.session)
    df_employee = (
        pyspark_employee.groupBy(pyspark_employee.age)
        .agg(F.min(pyspark_employee.employee_id).alias("min_employee_id"))
        .where(F.col("min_employee_id") > F.lit(1))
    )
    dfs_employee = (
        employee.groupBy(employee.age)
        .agg(min(employee.employee_id).alias("min_employee_id"))
        .where(col("min_employee_id") > lit(1))
    )
    compare_frames(df_employee, dfs_employee)


def test_cube(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    get_func: t.Callable,
    is_postgres: t.Callable,
    is_bigquery: t.Callable,
):
    employee = get_df("employee")
    max = get_func("max", employee.session)
    spark_agg_columns = [F.max(pyspark_employee.fname).alias("last_fname")]
    sqlf_agg_columns = [max(employee.fname).alias("last_fname")]
    if not is_postgres() and not is_bigquery():
        grouping_id = get_func("grouping_id", employee.session)
        spark_agg_columns.append(F.grouping_id().alias("g_id"))
        sqlf_agg_columns.append(grouping_id().alias("g_id"))
    df_employee = (
        pyspark_employee.cube("age", "store_id").agg(*spark_agg_columns).orderBy("age", "store_id")
    )
    dfs_employee = (
        employee.cube("age", "store_id").agg(*sqlf_agg_columns).orderBy("age", "store_id")
    )
    compare_frames(df_employee, dfs_employee)


def test_count(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.groupBy(pyspark_employee.age).count()
    dfs = employee.groupBy(employee.age).count()
    compare_frames(df, dfs)


def test_mean(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    is_redshift: t.Callable,
):
    if is_redshift():
        pytest.skip("Redshift will return an int for avg since the inputs are ints")
    employee = get_df("employee")
    df = pyspark_employee.groupBy().mean("age", "store_id")
    dfs = employee.groupBy().mean("age", "store_id")
    compare_frames(df, dfs)


def test_avg(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
    is_redshift: t.Callable,
):
    if is_redshift():
        pytest.skip("Redshift will return an int for avg since the inputs are ints")
    employee = get_df("employee")
    df = pyspark_employee.groupBy("age").avg("store_id")
    dfs = employee.groupBy("age").avg("store_id")
    compare_frames(df, dfs)


def test_max(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.groupBy("age").max("store_id")
    dfs = employee.groupBy("age").max("store_id")
    compare_frames(df, dfs)


def test_min(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.groupBy("age").min("store_id")
    dfs = employee.groupBy("age").min("store_id")
    compare_frames(df, dfs)


def test_sum(
    pyspark_employee: DataFrame,
    compare_frames: t.Callable,
    get_df: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.groupBy("age").sum("store_id")
    dfs = employee.groupBy("age").sum("store_id")
    compare_frames(df, dfs)



from __future__ import annotations

import logging
import sys
import typing as t

from sqlframe.base.dataframe import (
    _BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.base.mixins.dataframe_mixins import (
    NoCachePersistSupportMixin,
    TypedColumnsFromTempViewMixin,
)
from sqlframe.duckdb.group import DuckDBGroupedData

if t.TYPE_CHECKING:
    from sqlframe.duckdb.session import DuckDBSession  # noqa
    from sqlframe.duckdb.readwriter import DuckDBDataFrameWriter  # noqa
    from sqlframe.duckdb.group import DuckDBGroupedData  # noqa
    from pyarrow import Table as ArrowTable


logger = logging.getLogger(__name__)


class DuckDBDataFrameNaFunctions(_BaseDataFrameNaFunctions["DuckDBDataFrame"]):
    pass


class DuckDBDataFrameStatFunctions(_BaseDataFrameStatFunctions["DuckDBDataFrame"]):
    pass


class DuckDBDataFrame(
    NoCachePersistSupportMixin,
    TypedColumnsFromTempViewMixin,
    _BaseDataFrame[
        "DuckDBSession",
        "DuckDBDataFrameWriter",
        "DuckDBDataFrameNaFunctions",
        "DuckDBDataFrameStatFunctions",
        "DuckDBGroupedData",
    ],
):
    _na = DuckDBDataFrameNaFunctions
    _stat = DuckDBDataFrameStatFunctions
    _group_data = DuckDBGroupedData

    def toArrow(self) -> ArrowTable:
        self._collect(skip_rows=True)
        return self.session._last_result.arrow()



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import logging
import typing as t

from sqlglot import exp
from sqlglot.helper import ensure_list

from sqlframe.base.readerwriter import _BaseDataFrameReader, _BaseDataFrameWriter
from sqlframe.base.util import ensure_column_mapping, to_csv

if t.TYPE_CHECKING:
    from sqlframe.base._typing import OptionalPrimitiveType, PathOrPaths
    from sqlframe.base.types import StructType
    from sqlframe.duckdb.dataframe import DuckDBDataFrame
    from sqlframe.duckdb.session import DuckDBSession  # noqa

logger = logging.getLogger(__name__)


class DuckDBDataFrameReader(_BaseDataFrameReader["DuckDBSession", "DuckDBDataFrame"]):
    def load(
        self,
        path: t.Optional[PathOrPaths] = None,
        format: t.Optional[str] = None,
        schema: t.Optional[t.Union[StructType, str]] = None,
        **options: OptionalPrimitiveType,
    ) -> DuckDBDataFrame:
        """Loads data from a data source and returns it as a :class:`DataFrame`.

        .. versionadded:: 1.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        path : str or list, t.Optional
            t.Optional string or a list of string for file-system backed data sources.
        format : str, t.Optional
            t.Optional string for format of the data source. Default to 'parquet'.
        schema : :class:`pyspark.sql.types.StructType` or str, t.Optional
            t.Optional :class:`pyspark.sql.types.StructType` for the input schema
            or a DDL-formatted string (For example ``col0 INT, col1 DOUBLE``).
        **options : dict
            all other string options

        Examples
        --------
        Load a CSV file with format, schema and options specified.

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as d:
        ...     # Write a DataFrame into a CSV file with a header
        ...     df = spark.createDataFrame([{"age": 100, "name": "Hyukjin Kwon"}])
        ...     df.write.option("header", True).mode("overwrite").format("csv").save(d)
        ...
        ...     # Read the CSV file as a DataFrame with 'nullValue' option set to 'Hyukjin Kwon',
        ...     # and 'header' option set to `True`.
        ...     df = spark.read.load(
        ...         d, schema=df.schema, format="csv", nullValue="Hyukjin Kwon", header=True)
        ...     df.printSchema()
        ...     df.show()
        root
         |-- age: long (nullable = true)
         |-- name: string (nullable = true)
        +---+----+
        |age|name|
        +---+----+
        |100|NULL|
        +---+----+
        """
        if schema:
            column_mapping = ensure_column_mapping(schema)
            select_column_mapping = column_mapping.copy()
            if options.get("filename"):
                select_column_mapping["filename"] = "VARCHAR"
            select_columns = [x.expression for x in self._to_casted_columns(select_column_mapping)]
            if format == "csv":
                duckdb_columns = ", ".join(
                    [f"'{column}': '{dtype}'" for column, dtype in column_mapping.items()]
                )
                options["columns"] = "{" + duckdb_columns + "}"
        else:
            select_columns = [exp.Star()]
        if format:
            paths = ",".join([f"'{path}'" for path in ensure_list(path)])
            from_clause = f"read_{format}([{paths}], {to_csv(options)})"
        else:
            from_clause = f"'{path}'"
        df = self.session.sql(exp.select(*select_columns).from_(from_clause), qualify=False)
        self.session._last_loaded_file = path  # type: ignore
        return df


class DuckDBDataFrameWriter(_BaseDataFrameWriter["DuckDBSession", "DuckDBDataFrame"]):
    def _write(self, path: str, mode: t.Optional[str], **options):  # type: ignore
        mode, skip = self._validate_mode(path, mode)
        if skip:
            return
        if mode == "append":
            raise NotImplementedError("Append mode not supported")
        options = to_csv(options, equality_char=" ")  # type: ignore
        expressions = self._df._get_expressions()
        for i, expression in enumerate(expressions):
            if i < len(expressions) - 1:
                self._df.session._collect(expressions)
            else:
                sql = self._df.session._to_sql(expression)
                self._df.session._collect(f"COPY ({sql}) TO '{path}' ({options})")



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t
from decimal import Decimal

from sqlframe.base.exceptions import RowError


class DataType:
    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: t.Any) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other: t.Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return self.simpleString()

    @classmethod
    def typeName(cls) -> str:
        return cls.__name__[:-4].lower()

    def simpleString(self) -> str:
        return self.typeName()

    def jsonValue(self) -> t.Union[str, t.Dict[str, t.Any]]:
        return str(self)


class DataTypeWithLength(DataType):
    def __init__(self, length: int):
        self.length = length

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.length})"

    def __str__(self) -> str:
        return f"{self.typeName()}({self.length})"


class StringType(DataType):
    pass


class CharType(DataTypeWithLength):
    pass


class VarcharType(DataTypeWithLength):
    pass


class BinaryType(DataType):
    pass


class BooleanType(DataType):
    pass


class DateType(DataType):
    pass


class TimestampType(DataType):
    pass


class TimestampNTZType(DataType):
    @classmethod
    def typeName(cls) -> str:
        return "timestamp_ntz"


class DecimalType(DataType):
    def __init__(self, precision: int = 10, scale: int = 0):
        self.precision = precision
        self.scale = scale

    def simpleString(self) -> str:
        return f"decimal({self.precision}, {self.scale})"

    def jsonValue(self) -> str:
        return f"decimal({self.precision}, {self.scale})"

    def __repr__(self) -> str:
        return f"DecimalType({self.precision}, {self.scale})"


class DoubleType(DataType):
    pass


class FloatType(DataType):
    pass


class ByteType(DataType):
    def __str__(self) -> str:
        return "tinyint"


class IntegerType(DataType):
    def __str__(self) -> str:
        return "int"


class LongType(DataType):
    def __str__(self) -> str:
        return "bigint"


class ShortType(DataType):
    def __str__(self) -> str:
        return "smallint"


class ArrayType(DataType):
    def __init__(self, elementType: DataType, containsNull: bool = True):
        self.elementType = elementType
        self.containsNull = containsNull

    def __repr__(self) -> str:
        return f"ArrayType({self.elementType, str(self.containsNull)}"

    def simpleString(self) -> str:
        return f"array<{self.elementType.simpleString()}>"

    def jsonValue(self) -> t.Dict[str, t.Any]:
        return {
            "type": self.typeName(),
            "elementType": self.elementType.jsonValue(),
            "containsNull": self.containsNull,
        }


class MapType(DataType):
    def __init__(self, keyType: DataType, valueType: DataType, valueContainsNull: bool = True):
        self.keyType = keyType
        self.valueType = valueType
        self.valueContainsNull = valueContainsNull

    def __repr__(self) -> str:
        return f"MapType({self.keyType}, {self.valueType}, {str(self.valueContainsNull)})"

    def simpleString(self) -> str:
        return f"map<{self.keyType.simpleString()}, {self.valueType.simpleString()}>"

    def jsonValue(self) -> t.Dict[str, t.Any]:
        return {
            "type": self.typeName(),
            "keyType": self.keyType.jsonValue(),
            "valueType": self.valueType.jsonValue(),
            "valueContainsNull": self.valueContainsNull,
        }


class StructField(DataType):
    def __init__(
        self,
        name: str,
        dataType: DataType,
        nullable: bool = True,
        metadata: t.Optional[t.Dict[str, t.Any]] = None,
    ):
        self.name = name
        self.dataType = dataType
        self.nullable = nullable
        self.metadata = metadata or {}

    def __repr__(self) -> str:
        return f"StructField('{self.name}', {self.dataType}, {str(self.nullable)})"

    def simpleString(self) -> str:
        return f"{self.name}:{self.dataType.simpleString()}"

    def jsonValue(self) -> t.Dict[str, t.Any]:
        return {
            "name": self.name,
            "type": self.dataType.jsonValue(),
            "nullable": self.nullable,
            "metadata": self.metadata,
        }


class StructType(DataType):
    def __init__(self, fields: t.Optional[t.List[StructField]] = None):
        if not fields:
            self.fields = []
            self.names = []
        else:
            self.fields = fields
            self.names = [f.name for f in fields]

    def __iter__(self) -> t.Iterator[StructField]:
        return iter(self.fields)

    def __len__(self) -> int:
        return len(self.fields)

    def __repr__(self) -> str:
        return f"StructType({', '.join(str(field) for field in self)})"

    def simpleString(self) -> str:
        return f"struct<{', '.join(x.simpleString() for x in self)}>"

    def jsonValue(self) -> t.Dict[str, t.Any]:
        return {"type": self.typeName(), "fields": [x.jsonValue() for x in self]}

    def fieldNames(self) -> t.List[str]:
        return list(self.names)


def _create_row(
    fields: t.Union[Row, t.List[str]], values: t.Union[t.Tuple[t.Any, ...], t.List[t.Any]]
) -> Row:
    row = Row(*[float(x) if isinstance(x, Decimal) else x for x in values])
    row.__fields__ = fields
    return row


class Row(tuple):
    """
    A row in :class:`DataFrame`.
    The fields in it can be accessed:

    * like attributes (``row.key``)
    * like dictionary values (``row[key]``)

    ``key in row`` will search through row keys.

    Row can be used to create a row object by using named arguments.
    It is not allowed to omit a named argument to represent that the value is
    None or missing. This should be explicitly set to None in this case.

    .. versionchanged:: 3.0.0
        Rows created from named arguments no longer have
        field names sorted alphabetically and will be ordered in the position as
        entered.

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> row = Row(name="Alice", age=11)
    >>> row
    Row(name='Alice', age=11)
    >>> row['name'], row['age']
    ('Alice', 11)
    >>> row.name, row.age
    ('Alice', 11)
    >>> 'name' in row
    True
    >>> 'wrong_key' in row
    False

    Row also can be used to create another Row like class, then it
    could be used to create Row objects, such as

    >>> Person = Row("name", "age")
    >>> Person
    <Row('name', 'age')>
    >>> 'name' in Person
    True
    >>> 'wrong_key' in Person
    False
    >>> Person("Alice", 11)
    Row(name='Alice', age=11)

    This form can also be used to create rows as tuple values, i.e. with unnamed
    fields.

    >>> row1 = Row("Alice", 11)
    >>> row2 = Row(name="Alice", age=11)
    >>> row1 == row2
    True
    """

    @t.overload
    def __new__(cls, *args: str) -> Row: ...

    @t.overload
    def __new__(cls, **kwargs: t.Any) -> Row: ...

    def __new__(cls, *args: t.Optional[str], **kwargs: t.Optional[t.Any]) -> Row:
        if args and kwargs:
            raise RowError("Cannot use both args and kwargs to create Row")
        if kwargs:
            # create row objects
            # psycopg2 returns Decimal type for numeric while PySpark returns float so we convert to float
            row = tuple.__new__(
                cls, [float(x) if isinstance(x, Decimal) else x for x in kwargs.values()]
            )
            row.__fields__ = list(kwargs.keys())
            return row
        else:
            # create row class or objects
            return tuple.__new__(cls, args)

    def asDict(self, recursive: bool = False) -> t.Dict[str, t.Any]:
        """
        Return as a dict

        Parameters
        ----------
        recursive : bool, optional
            turns the nested Rows to dict (default: False).

        Notes
        -----
        If a row contains duplicate field names, e.g., the rows of a join
        between two :class:`DataFrame` that both have the fields of same names,
        one of the duplicate fields will be selected by ``asDict``. ``__getitem__``
        will also return one of the duplicate fields, however returned value might
        be different to ``asDict``.

        Examples
        --------
        >>> from pyspark.sql import Row
        >>> Row(name="Alice", age=11).asDict() == {'name': 'Alice', 'age': 11}
        True
        >>> row = Row(key=1, value=Row(name='a', age=2))
        >>> row.asDict() == {'key': 1, 'value': Row(name='a', age=2)}
        True
        >>> row.asDict(True) == {'key': 1, 'value': {'name': 'a', 'age': 2}}
        True
        """
        if not hasattr(self, "__fields__"):
            raise RowError("Cannot convert a Row class into dict")

        if recursive:

            def conv(obj: t.Any) -> t.Any:
                if isinstance(obj, Row):
                    return obj.asDict(True)
                elif isinstance(obj, list):
                    return [conv(o) for o in obj]
                elif isinstance(obj, dict):
                    return dict((k, conv(v)) for k, v in obj.items())
                else:
                    return obj

            return dict(zip(self.__fields__, (conv(o) for o in self)))
        else:
            return dict(zip(self.__fields__, self))

    def __contains__(self, item: t.Any) -> bool:
        if hasattr(self, "__fields__"):
            return item in self.__fields__
        else:
            return super(Row, self).__contains__(item)

    # let object acts like class
    def __call__(self, *args: t.Any) -> Row:
        """create new Row object"""
        if len(args) > len(self):
            raise RowError(
                "Cannot create Row with fields %s. Expected %d values but got %s instead"
                % (self, len(self), args)
            )

        return _create_row(self, args)

    def __getitem__(self, item: t.Any) -> t.Any:
        if isinstance(item, (int, slice)):
            return super(Row, self).__getitem__(item)
        try:
            # it will be slow when it has many fields,
            # but this will not be used in normal cases
            idx = self.__fields__.index(item)
            return super(Row, self).__getitem__(idx)
        except IndexError:
            raise KeyError(item)
        except ValueError:
            raise RowError(item)

    def __getattr__(self, item: str) -> t.Any:
        if item.startswith("__"):
            raise AttributeError(item)
        try:
            # it will be slow when it has many fields,
            # but this will not be used in normal cases
            idx = self.__fields__.index(item)
            return self[idx]
        except IndexError:
            raise AttributeError(item)
        except ValueError:
            raise AttributeError(item)

    def __setattr__(self, key: t.Any, value: t.Any) -> None:
        if key != "__fields__":
            raise RuntimeError("Row is read-only")
        self.__dict__[key] = value

    def __reduce__(
        self,
    ) -> t.Union[str, t.Tuple[t.Any, ...]]:
        """Returns a tuple so Python knows how to pickle Row."""
        if hasattr(self, "__fields__"):
            return (_create_row, (self.__fields__, tuple(self)))
        else:
            return tuple.__reduce__(self)

    def __repr__(self) -> str:
        """Printable representation of Row used in Python REPL."""
        if hasattr(self, "__fields__"):
            return "Row(%s)" % ", ".join(
                "%s=%r" % (k, v) for k, v in zip(self.__fields__, tuple(self))
            )
        else:
            return "<Row(%s)>" % ", ".join(repr(field) for field in self)

    # SQLFrame Specific
    @property
    def _unique_field_names(self) -> t.List[str]:
        fields = []
        for i, field in enumerate(self.__fields__):
            if field in fields:
                field = field + "_" + str(i)
            fields.append(field)
        return fields



from sqlframe.spark.catalog import SparkCatalog
from sqlframe.spark.column import Column
from sqlframe.spark.dataframe import SparkDataFrame, SparkDataFrameNaFunctions
from sqlframe.spark.group import SparkGroupedData
from sqlframe.spark.readwriter import (
    SparkDataFrameReader,
    SparkDataFrameWriter,
)
from sqlframe.spark.session import SparkSession
from sqlframe.spark.window import Window, WindowSpec

__all__ = [
    "SparkCatalog",
    "Column",
    "SparkDataFrame",
    "SparkDataFrameNaFunctions",
    "SparkGroupedData",
    "SparkDataFrameReader",
    "SparkDataFrameWriter",
    "SparkSession",
    "Window",
    "WindowSpec",
]



from __future__ import annotations

import typing as t

from sqlglot import exp
from sqlglot.helper import classproperty, ensure_list

from sqlframe.base.session import _BaseSession
from sqlframe.base.util import normalize_string
from sqlframe.spark.catalog import SparkCatalog
from sqlframe.spark.dataframe import SparkDataFrame
from sqlframe.spark.readwriter import (
    SparkDataFrameReader,
    SparkDataFrameWriter,
)
from sqlframe.spark.types import Row
from sqlframe.spark.udf import SparkUDFRegistration

if t.TYPE_CHECKING:
    import pandas as pd
    from pyspark.sql.session import SparkConf
    from pyspark.sql.session import SparkSession as PySparkSession

    from sqlframe.base._typing import OptionalPrimitiveType
else:
    PySparkSession = t.Any


class SparkSession(
    _BaseSession[  # type: ignore
        SparkCatalog,
        SparkDataFrameReader,
        SparkDataFrameWriter,
        SparkDataFrame,
        PySparkSession,
        SparkUDFRegistration,
    ],
):
    _catalog = SparkCatalog
    _reader = SparkDataFrameReader
    _writer = SparkDataFrameWriter
    _df = SparkDataFrame
    _udf_registration = SparkUDFRegistration

    def __init__(self, conn: t.Optional[PySparkSession] = None, *args, **kwargs):
        from pyspark.sql.session import DataFrame, SparkSession

        self._last_df: t.Optional[DataFrame] = None
        if not hasattr(self, "spark_session"):
            super().__init__(conn)
            self.spark_session = conn or SparkSession.builder.getOrCreate()

    @property
    def _conn(self) -> t.Any:
        raise NotImplementedError()

    @property
    def _cur(self) -> t.Any:
        raise NotImplementedError()

    def _collect(
        self,
        expressions: t.Union[str, exp.Expression, t.List[str], t.List[exp.Expression]],
        *,
        quote_identifiers: bool = True,
        skip_normalization: bool = False,
        skip_rows: bool = False,
    ) -> t.List[Row]:
        for expression in ensure_list(expressions):
            sql = (
                self._to_sql(expression, quote_identifiers=quote_identifiers)
                if isinstance(expression, exp.Expression)
                else expression
            )
            self._execute(sql)  # type: ignore
        if skip_rows:
            return []
        assert self._last_df is not None
        return [
            Row(
                **{
                    normalize_string(
                        k, from_dialect="execution", to_dialect="output", is_column=True
                    ): v
                    for k, v in row.asDict().items()
                }
            )
            for row in self._last_df.collect()
        ]

    def _execute(self, sql: str) -> None:
        self._last_df = self.spark_session.sql(sql)

    def _fetchdf(
        self,
        expressions: t.Union[exp.Expression, t.List[exp.Expression]],
        *,
        quote_identifiers: bool = True,
    ) -> pd.DataFrame:
        for expression in ensure_list(expressions):
            sql = (
                self._to_sql(expression, quote_identifiers=quote_identifiers)
                if isinstance(expression, exp.Expression)
                else expression
            )
            self._execute(sql)  # type: ignore
        assert self._last_df is not None
        return self._last_df.toPandas()

    @property
    def _has_connection(self) -> bool:
        return True

    @classproperty
    def builder(cls) -> Builder:  # type: ignore
        """Creates a :class:`Builder` for constructing a :class:`SparkSession`.

        .. versionchanged:: 3.4.0
            Supports Spark Connect.
        """
        return cls.Builder()

    class Builder(_BaseSession.Builder):
        def __init__(self):
            from pyspark.sql.session import SparkSession

            super().__init__()
            self.builder = SparkSession.builder

        def config(
            self,
            key: t.Optional[str] = None,
            value: t.Optional[t.Any] = None,
            conf: t.Optional[SparkConf] = None,
            *,
            map: t.Optional[t.Dict[str, OptionalPrimitiveType]] = None,
        ) -> SparkSession.Builder:
            super().config(key, value, map=map)
            self.builder = self.builder.config(key, value, conf, map=map)
            return self

        def master(self, master: str) -> SparkSession.Builder:
            self.builder = self.builder.config("spark.master", master)
            return self

        def remote(self, url: str) -> SparkSession.Builder:
            self.builder = self.builder.config("spark.remote", url)
            return self

        def appName(self, name: str) -> SparkSession.Builder:
            self.builder = self.builder.appName(name)
            return self

        def enableHiveSupport(self) -> SparkSession.Builder:
            self.builder = self.builder.enableHiveSupport()
            return self

        @property
        def session(self) -> SparkSession:
            if "conn" not in self._session_kwargs:
                self._session_kwargs["conn"] = self.builder.getOrCreate()
            return SparkSession(**self._session_kwargs)

        def getOrCreate(self) -> SparkSession:
            self._set_session_properties()
            return self.session



from sqlframe.base.function_alternatives import (  # noqa
    percentile_without_disc as percentile,
    add_months_by_multiplication as add_months,
    arrays_overlap_renamed as arrays_overlap,
)
from sqlframe.base.functions import (
    abs as abs,
    acos as acos,
    acosh as acosh,
    aes_decrypt as aes_decrypt,
    aes_encrypt as aes_encrypt,
    aggregate as aggregate,
    any_value as any_value,
    approxCountDistinct as approxCountDistinct,
    approx_count_distinct as approx_count_distinct,
    approx_percentile as approx_percentile,
    array as array,
    array_agg as array_agg,
    array_append as array_append,
    array_compact as array_compact,
    array_contains as array_contains,
    array_distinct as array_distinct,
    array_except as array_except,
    array_insert as array_insert,
    array_intersect as array_intersect,
    array_join as array_join,
    array_max as array_max,
    array_min as array_min,
    array_position as array_position,
    array_prepend as array_prepend,
    array_remove as array_remove,
    array_repeat as array_repeat,
    array_size as array_size,
    array_sort as array_sort,
    array_union as array_union,
    arrays_zip as arrays_zip,
    asc as asc,
    asc_nulls_first as asc_nulls_first,
    asc_nulls_last as asc_nulls_last,
    ascii as ascii,
    asin as asin,
    asinh as asinh,
    assert_true as assert_true,
    atan as atan,
    atan2 as atan2,
    atanh as atanh,
    avg as avg,
    base64 as base64,
    bin as bin,
    bit_and as bit_and,
    bit_count as bit_count,
    bit_get as bit_get,
    bit_length as bit_length,
    bit_or as bit_or,
    bit_xor as bit_xor,
    bitmap_bit_position as bitmap_bit_position,
    bitmap_bucket_number as bitmap_bucket_number,
    bitmap_construct_agg as bitmap_construct_agg,
    bitmap_count as bitmap_count,
    bitmap_or_agg as bitmap_or_agg,
    bitwiseNOT as bitwiseNOT,
    bitwise_not as bitwise_not,
    bool_and as bool_and,
    bool_or as bool_or,
    broadcast as broadcast,
    bround as bround,
    btrim as btrim,
    bucket as bucket,
    call_function as call_function,
    cardinality as cardinality,
    cbrt as cbrt,
    ceil as ceil,
    ceiling as ceiling,
    char as char,
    char_length as char_length,
    character_length as character_length,
    coalesce as coalesce,
    col as col,
    collect_list as collect_list,
    collect_set as collect_set,
    concat as concat,
    concat_ws as concat_ws,
    contains as contains,
    conv as conv,
    convert_timezone as convert_timezone,
    corr as corr,
    cos as cos,
    cosh as cosh,
    cot as cot,
    count as count,
    countDistinct as countDistinct,
    count_distinct as count_distinct,
    count_if as count_if,
    count_min_sketch as count_min_sketch,
    covar_pop as covar_pop,
    covar_samp as covar_samp,
    crc32 as crc32,
    create_map as create_map,
    csc as csc,
    cume_dist as cume_dist,
    curdate as curdate,
    current_catalog as current_catalog,
    current_database as current_database,
    current_date as current_date,
    current_schema as current_schema,
    current_timestamp as current_timestamp,
    current_timezone as current_timezone,
    current_user as current_user,
    date_add as date_add,
    date_diff as date_diff,
    date_format as date_format,
    date_from_unix_date as date_from_unix_date,
    date_part as date_part,
    date_sub as date_sub,
    date_trunc as date_trunc,
    dateadd as dateadd,
    datediff as datediff,
    datepart as datepart,
    day as day,
    dayofmonth as dayofmonth,
    dayofweek as dayofweek,
    dayofyear as dayofyear,
    days as days,
    decode as decode,
    degrees as degrees,
    dense_rank as dense_rank,
    desc as desc,
    desc_nulls_first as desc_nulls_first,
    desc_nulls_last as desc_nulls_last,
    e as e,
    element_at as element_at,
    elt as elt,
    encode as encode,
    endswith as endswith,
    equal_null as equal_null,
    every as every,
    exists as exists,
    exp as exp,
    explode as explode,
    explode_outer as explode_outer,
    expm1 as expm1,
    expr as expr,
    extract as extract,
    factorial as factorial,
    filter as filter,
    find_in_set as find_in_set,
    first as first,
    first_value as first_value,
    flatten as flatten,
    floor as floor,
    forall as forall,
    format_number as format_number,
    format_string as format_string,
    from_csv as from_csv,
    from_json as from_json,
    from_unixtime as from_unixtime,
    from_utc_timestamp as from_utc_timestamp,
    get as get,
    get_active_spark_context as get_active_spark_context,
    get_json_object as get_json_object,
    getbit as getbit,
    greatest as greatest,
    grouping as grouping,
    grouping_id as grouping_id,
    hash as hash,
    hex as hex,
    histogram_numeric as histogram_numeric,
    hll_sketch_agg as hll_sketch_agg,
    hll_sketch_estimate as hll_sketch_estimate,
    hll_union as hll_union,
    hll_union_agg as hll_union_agg,
    hour as hour,
    hours as hours,
    hypot as hypot,
    ifnull as ifnull,
    ilike as ilike,
    initcap as initcap,
    inline as inline,
    inline_outer as inline_outer,
    input_file_name as input_file_name,
    instr as instr,
    isnan as isnan,
    isnotnull as isnotnull,
    isnull as isnull,
    java_method as java_method,
    json_array_length as json_array_length,
    json_object_keys as json_object_keys,
    json_tuple as json_tuple,
    kurtosis as kurtosis,
    lag as lag,
    last as last,
    last_day as last_day,
    last_value as last_value,
    lcase as lcase,
    lead as lead,
    least as least,
    left as left,
    length as length,
    levenshtein as levenshtein,
    like as like,
    lit as lit,
    ln as ln,
    localtimestamp as localtimestamp,
    locate as locate,
    log as log,
    log10 as log10,
    log1p as log1p,
    log2 as log2,
    lower as lower,
    lpad as lpad,
    ltrim as ltrim,
    make_date as make_date,
    make_dt_interval as make_dt_interval,
    make_interval as make_interval,
    make_timestamp as make_timestamp,
    make_timestamp_ltz as make_timestamp_ltz,
    make_timestamp_ntz as make_timestamp_ntz,
    make_ym_interval as make_ym_interval,
    map_concat as map_concat,
    map_contains_key as map_contains_key,
    map_entries as map_entries,
    map_filter as map_filter,
    map_from_arrays as map_from_arrays,
    map_from_entries as map_from_entries,
    map_keys as map_keys,
    map_values as map_values,
    map_zip_with as map_zip_with,
    mask as mask,
    max as max,
    max_by as max_by,
    md5 as md5,
    mean as mean,
    median as median,
    min as min,
    min_by as min_by,
    minute as minute,
    mode as mode,
    monotonically_increasing_id as monotonically_increasing_id,
    month as month,
    months as months,
    months_between as months_between,
    named_struct as named_struct,
    nanvl as nanvl,
    negate as negate,
    negative as negative,
    next_day as next_day,
    now as now,
    nth_value as nth_value,
    ntile as ntile,
    nullif as nullif,
    nvl as nvl,
    nvl2 as nvl2,
    octet_length as octet_length,
    overlay as overlay,
    parse_url as parse_url,
    percent_rank as percent_rank,
    percentile_approx as percentile_approx,
    pi as pi,
    pmod as pmod,
    posexplode as posexplode,
    posexplode_outer as posexplode_outer,
    position as position,
    positive as positive,
    pow as pow,
    power as power,
    printf as printf,
    quarter as quarter,
    radians as radians,
    raise_error as raise_error,
    rand as rand,
    randn as randn,
    rank as rank,
    reduce as reduce,
    reflect as reflect,
    regexp as regexp,
    regexp_count as regexp_count,
    regexp_extract as regexp_extract,
    regexp_extract_all as regexp_extract_all,
    regexp_instr as regexp_instr,
    regexp_like as regexp_like,
    regexp_replace as regexp_replace,
    regexp_substr as regexp_substr,
    regr_avgx as regr_avgx,
    regr_avgy as regr_avgy,
    regr_count as regr_count,
    regr_intercept as regr_intercept,
    regr_r2 as regr_r2,
    regr_slope as regr_slope,
    regr_sxx as regr_sxx,
    regr_sxy as regr_sxy,
    regr_syy as regr_syy,
    repeat as repeat,
    replace as replace,
    reverse as reverse,
    right as right,
    rint as rint,
    rlike as rlike,
    round as round,
    row_number as row_number,
    rpad as rpad,
    rtrim as rtrim,
    schema_of_csv as schema_of_csv,
    schema_of_json as schema_of_json,
    sec as sec,
    second as second,
    sentences as sentences,
    sequence as sequence,
    sha as sha,
    sha1 as sha1,
    sha2 as sha2,
    shiftLeft as shiftLeft,
    shiftRight as shiftRight,
    shiftRightUnsigned as shiftRightUnsigned,
    shiftleft as shiftleft,
    shiftright as shiftright,
    shiftrightunsigned as shiftrightunsigned,
    shuffle as shuffle,
    sign as sign,
    signum as signum,
    sin as sin,
    sinh as sinh,
    size as size,
    skewness as skewness,
    slice as slice,
    some as some,
    sort_array as sort_array,
    soundex as soundex,
    spark_partition_id as spark_partition_id,
    split as split,
    split_part as split_part,
    sqrt as sqrt,
    stack as stack,
    startswith as startswith,
    std as std,
    stddev as stddev,
    stddev_pop as stddev_pop,
    stddev_samp as stddev_samp,
    str_to_map as str_to_map,
    struct as struct,
    substr as substr,
    substring as substring,
    substring_index as substring_index,
    sum as sum,
    sumDistinct as sumDistinct,
    sum_distinct as sum_distinct,
    tan as tan,
    tanh as tanh,
    timestamp_micros as timestamp_micros,
    timestamp_millis as timestamp_millis,
    timestamp_seconds as timestamp_seconds,
    toDegrees as toDegrees,
    toRadians as toRadians,
    to_binary as to_binary,
    to_char as to_char,
    to_csv as to_csv,
    to_date as to_date,
    to_json as to_json,
    to_number as to_number,
    to_timestamp as to_timestamp,
    to_timestamp_ltz as to_timestamp_ltz,
    to_timestamp_ntz as to_timestamp_ntz,
    to_unix_timestamp as to_unix_timestamp,
    to_utc_timestamp as to_utc_timestamp,
    to_varchar as to_varchar,
    transform as transform,
    transform_keys as transform_keys,
    transform_values as transform_values,
    translate as translate,
    trim as trim,
    trunc as trunc,
    try_add as try_add,
    try_aes_decrypt as try_aes_decrypt,
    try_avg as try_avg,
    try_divide as try_divide,
    try_element_at as try_element_at,
    try_multiply as try_multiply,
    try_subtract as try_subtract,
    try_sum as try_sum,
    try_to_binary as try_to_binary,
    try_to_number as try_to_number,
    try_to_timestamp as try_to_timestamp,
    typeof as typeof,
    ucase as ucase,
    unbase64 as unbase64,
    unhex as unhex,
    unix_date as unix_date,
    unix_micros as unix_micros,
    unix_millis as unix_millis,
    unix_seconds as unix_seconds,
    unix_timestamp as unix_timestamp,
    upper as upper,
    url_decode as url_decode,
    url_encode as url_encode,
    user as user,
    var_pop as var_pop,
    var_samp as var_samp,
    variance as variance,
    version as version,
    weekday as weekday,
    weekofyear as weekofyear,
    when as when,
    width_bucket as width_bucket,
    xpath as xpath,
    xpath_boolean as xpath_boolean,
    xpath_double as xpath_double,
    xpath_float as xpath_float,
    xpath_int as xpath_int,
    xpath_long as xpath_long,
    xpath_number as xpath_number,
    xpath_short as xpath_short,
    xpath_string as xpath_string,
    xxhash64 as xxhash64,
    year as year,
    years as years,
    zip_with as zip_with,
)



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

import typing as t

from sqlglot import expressions as exp


def replace_id_value(
    node: exp.Expression, replacement_mapping: t.Dict[exp.Identifier, exp.Identifier]
) -> exp.Expression:
    if isinstance(node, exp.Identifier) and node in replacement_mapping:
        node = node.replace(replacement_mapping[node].copy())
    return node



import inspect
import sys

import sqlframe.base.functions

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines")
        and "postgres" not in func.unsupported_engines
        and "*" not in func.unsupported_engines
    }
)


from sqlframe.base.function_alternatives import (  # noqa
    any_value_ignore_nulls_not_supported as any_value,
    e_literal as e,
    expm1_from_exp as expm1,
    log1p_from_log as log1p,
    rint_from_round as rint,
    collect_set_from_list_distinct as collect_set,
    isnan_using_equal as isnan,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    rand_no_seed as rand,
    round_cast_as_numeric as round,
    year_from_extract as year,
    quarter_from_extract as quarter,
    month_from_extract as month,
    dayofweek_from_extract_with_isodow as dayofweek,
    dayofmonth_from_extract_with_day as dayofmonth,
    dayofyear_from_extract_doy as dayofyear,
    hour_from_extract as hour,
    minute_from_extract as minute,
    second_from_extract as second,
    weekofyear_from_extract_as_week as weekofyear,
    make_date_casted_as_integer as make_date,
    date_add_by_multiplication as date_add,
    date_sub_by_multiplication as date_sub,
    date_diff_with_subtraction as date_diff,
    date_diff_with_subtraction as datediff,
    add_months_by_multiplication as add_months,
    months_between_from_age_and_extract as months_between,
    from_unixtime_from_timestamp as from_unixtime,
    unix_timestamp_from_extract as unix_timestamp,
    base64_from_blob as base64,
    bas64_from_encode as base64,
    unbase64_from_decode as unbase64,
    decode_from_convert_from as decode,
    encode_from_convert_to as encode,
    format_number_from_to_char as format_number,
    format_string_with_format as format_string,
    split_from_regex_split_to_array as split,
    array_contains_any as array_contains,
    slice_with_brackets as slice,
    element_at_using_brackets as element_at,
    get_json_object_using_arrow_op as get_json_object,
    array_min_from_subquery as array_min,
    array_max_from_subquery as array_max,
    left_cast_len as left,
    right_cast_len as right,
    position_cast_start as position,
    try_element_at_zero_based as try_element_at,
    try_to_timestamp_pgtemp as try_to_timestamp,
    typeof_pg_typeof as typeof,
    _is_string_using_typeof_char_varying as _is_string,
    endswith_using_like as endswith,
    last_day_with_cast as last_day,
)



from __future__ import annotations

import typing as t

import pytest
from sqlglot import Dialect, exp, parse_one

if t.TYPE_CHECKING:
    from sqlframe.base.session import _BaseSession

pytest_plugins = ["tests.integration.fixtures"]


@pytest.fixture
def cleanup_session(get_session: t.Callable[[], _BaseSession]) -> t.Iterator[_BaseSession]:
    session = get_session()
    yield session
    session._execute("DROP TABLE IF EXISTS test_table")


def test_session(cleanup_session: _BaseSession):
    session = cleanup_session
    session._execute("DROP TABLE IF EXISTS test_table")
    session._collect(
        parse_one(
            "CREATE TABLE test_table (cola INT, colb STRING, `col with space` STRING)",
            dialect="spark",
        )
    )
    columns = session.catalog.get_columns("test_table")
    if session.execution_dialect == Dialect.get_or_raise("bigquery"):
        cola_type = exp.DataType.build("INT64", dialect=session.execution_dialect)
    elif session.execution_dialect == Dialect.get_or_raise("snowflake"):
        cola_type = exp.DataType.build("DECIMAL", dialect=session.execution_dialect)
    else:
        cola_type = exp.DataType.build("INT", dialect=session.execution_dialect)
    assert columns == {
        "cola": cola_type,
        "colb": exp.DataType.build("STRING", dialect=session.output_dialect),
        "`col with space`": exp.DataType.build("STRING", dialect=session.output_dialect),
    }



from __future__ import annotations

import os
import typing as t

import pytest
from sqlglot import exp

from sqlframe.redshift.session import RedshiftSession

if t.TYPE_CHECKING:
    from redshift_connector.core import Connection as RedshiftConnection

pytest_plugins = ["tests.common_fixtures"]
pytestmark = [
    pytest.mark.redshift,
    pytest.mark.xdist_group("redshift_tests"),
]


@pytest.fixture
def cleanup_connector() -> t.Iterator[RedshiftConnection]:
    from redshift_connector import connect

    conn = connect(
        user=os.environ["SQLFRAME_REDSHIFT_USER"],
        password=os.environ["SQLFRAME_REDSHIFT_PASSWORD"],
        database=os.environ["SQLFRAME_REDSHIFT_DATABASE"],
        host=os.environ["SQLFRAME_REDSHIFT_HOST"],
        port=int(os.environ["SQLFRAME_REDSHIFT_PORT"]),
    )
    conn.autocommit = True
    conn.cursor().execute("CREATE SCHEMA IF NOT EXISTS db1")
    conn.cursor().execute("DROP TABLE IF EXISTS db1.test_table")
    conn.cursor().execute("CREATE TABLE IF NOT EXISTS db1.test_table (cola INT, colb VARCHAR(MAX))")

    yield conn
    conn.cursor().execute("DROP TABLE IF EXISTS db1.test_table")


def test_session_from_config(cleanup_connector: RedshiftConnection):
    session = RedshiftSession.builder.config("sqlframe.conn", cleanup_connector).getOrCreate()
    columns = session.catalog.get_columns("dev.db1.test_table")
    assert columns == {
        "cola": exp.DataType.build("INT", dialect=session.output_dialect),
        "colb": exp.DataType.build("STRING", dialect=session.output_dialect),
    }



from sqlframe.postgres.catalog import PostgresCatalog
from sqlframe.postgres.column import Column
from sqlframe.postgres.dataframe import PostgresDataFrame, PostgresDataFrameNaFunctions
from sqlframe.postgres.group import PostgresGroupedData
from sqlframe.postgres.readwriter import (
    PostgresDataFrameReader,
    PostgresDataFrameWriter,
)
from sqlframe.postgres.session import PostgresSession
from sqlframe.postgres.window import Window, WindowSpec

__all__ = [
    "PostgresCatalog",
    "Column",
    "PostgresDataFrame",
    "PostgresDataFrameNaFunctions",
    "PostgresGroupedData",
    "PostgresDataFrameReader",
    "PostgresDataFrameWriter",
    "PostgresSession",
    "Window",
    "WindowSpec",
]



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

from sqlframe.base.session import _BaseSession
from sqlframe.standalone.catalog import StandaloneCatalog
from sqlframe.standalone.dataframe import StandaloneDataFrame
from sqlframe.standalone.readwriter import (
    StandaloneDataFrameReader,
    StandaloneDataFrameWriter,
)
from sqlframe.standalone.udf import StandaloneUDFRegistration


class StandaloneSession(
    _BaseSession[  # type: ignore
        StandaloneCatalog,
        StandaloneDataFrameReader,
        StandaloneDataFrameWriter,
        StandaloneDataFrame,
        object,
        StandaloneUDFRegistration,
    ]
):  # type: ignore
    _catalog = StandaloneCatalog
    _reader = StandaloneDataFrameReader
    _writer = StandaloneDataFrameWriter
    _df = StandaloneDataFrame
    _udf_registration = StandaloneUDFRegistration

    class Builder(_BaseSession.Builder):
        DEFAULT_INPUT_DIALECT = "spark"
        DEFAULT_OUTPUT_DIALECT = "spark"

        @property
        def session(self) -> StandaloneSession:
            return StandaloneSession()

        def getOrCreate(self) -> StandaloneSession:
            self._set_session_properties()
            return self.session

    builder = Builder()



from __future__ import annotations

import typing as t

from sqlframe.base.types import Row

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import _BaseDataFrame

pytest_plugins = ["tests.integration.fixtures"]


def test_collect(get_engine_df: t.Callable[[str], _BaseDataFrame], get_func):
    employee = get_engine_df("employee")
    col = get_func("col", employee.session)
    results = employee.select(col("fname"), col("lname")).collect()
    assert results == [
        Row(**{"fname": "Jack", "lname": "Shephard"}),
        Row(**{"fname": "John", "lname": "Locke"}),
        Row(**{"fname": "Kate", "lname": "Austen"}),
        Row(**{"fname": "Claire", "lname": "Littleton"}),
        Row(**{"fname": "Hugo", "lname": "Reyes"}),
    ]


def test_show(
    get_engine_df: t.Callable[[str], _BaseDataFrame],
    get_func,
    capsys,
    caplog,
):
    employee = get_engine_df("employee")
    lit = get_func("lit", employee.session)
    employee = employee.select("*", lit(1).alias("one"))
    employee.show()
    captured = capsys.readouterr()
    assert (
        captured.out
        == """+-------------+--------+-----------+-----+----------+-----+
| employee_id | fname  |   lname   | age | store_id | one |
+-------------+--------+-----------+-----+----------+-----+
|      1      |  Jack  |  Shephard |  37 |    1     |  1  |
|      2      |  John  |   Locke   |  65 |    1     |  1  |
|      3      |  Kate  |   Austen  |  37 |    2     |  1  |
|      4      | Claire | Littleton |  27 |    2     |  1  |
|      5      |  Hugo  |   Reyes   |  29 |   100    |  1  |
+-------------+--------+-----------+-----+----------+-----+\n"""
    )
    assert "Truncate is ignored so full results will be displayed" not in caplog.text
    employee.show(truncate=True)
    captured = capsys.readouterr()
    assert "Truncate is ignored so full results will be displayed" in caplog.text


def test_show_limit(
    get_engine_df: t.Callable[[str], _BaseDataFrame], capsys, is_snowflake: t.Callable
):
    employee = get_engine_df("employee")
    employee.show(1)
    captured = capsys.readouterr()
    assert (
        captured.out
        == """+-------------+-------+----------+-----+----------+
| employee_id | fname |  lname   | age | store_id |
+-------------+-------+----------+-----+----------+
|      1      |  Jack | Shephard |  37 |    1     |
+-------------+-------+----------+-----+----------+\n"""
    )



from __future__ import annotations

import logging
import sys
import typing as t

from sqlframe.base.dataframe import (
    _BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.base.mixins.dataframe_mixins import NoCachePersistSupportMixin
from sqlframe.redshift.group import RedshiftGroupedData

if t.TYPE_CHECKING:
    from sqlframe.redshift.readwriter import RedshiftDataFrameWriter
    from sqlframe.redshift.session import RedshiftSession


logger = logging.getLogger(__name__)


class RedshiftDataFrameNaFunctions(_BaseDataFrameNaFunctions["RedshiftDataFrame"]):
    pass


class RedshiftDataFrameStatFunctions(_BaseDataFrameStatFunctions["RedshiftDataFrame"]):
    pass


class RedshiftDataFrame(
    NoCachePersistSupportMixin,
    _BaseDataFrame[
        "RedshiftSession",
        "RedshiftDataFrameWriter",
        "RedshiftDataFrameNaFunctions",
        "RedshiftDataFrameStatFunctions",
        "RedshiftGroupedData",
    ],
):
    _na = RedshiftDataFrameNaFunctions
    _stat = RedshiftDataFrameStatFunctions
    _group_data = RedshiftGroupedData



from __future__ import annotations

import inspect
import sys
import typing as t

from sqlglot import exp as sqlglot_expression

from sqlframe.base.util import (
    get_func_from_session,
)
from sqlframe.bigquery.column import Column

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrLiteral, ColumnOrName

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines")
        and "bigquery" not in func.unsupported_engines
        and "*" not in func.unsupported_engines
    }
)


from sqlframe.base.function_alternatives import (  # noqa
    any_value_ignore_nulls_not_supported as any_value,
    current_user_from_session_user as current_user,
    e_literal as e,
    expm1_from_exp as expm1,
    extract_convert_to_var as extract,
    factorial_from_case_statement as factorial,
    log1p_from_log as log1p,
    rint_from_round as rint,
    collect_set_from_list_distinct as collect_set,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    percentile_approx_without_accuracy_and_plural as percentile_approx,
    rand_no_seed as rand,
    year_from_extract as year,
    quarter_from_extract as quarter,
    month_from_extract as month,
    dayofweek_from_extract as dayofweek,
    dayofmonth_from_extract_with_day as dayofmonth,
    dayofyear_from_extract as dayofyear,
    hour_from_extract as hour,
    minute_from_extract as minute,
    second_from_extract as second,
    weekofyear_from_extract_as_isoweek as weekofyear,
    make_date_from_date_func as make_date,
    to_date_from_timestamp as to_date,
    last_day_with_cast as last_day,
    sha1_force_sha1_and_to_hex as sha,
    sha1_force_sha1_and_to_hex as sha1,
    hash_from_farm_fingerprint as hash,
    base64_from_blob as base64,
    concat_ws_from_array_to_string as concat_ws,
    format_string_with_format as format_string,
    instr_using_strpos as instr,
    overlay_from_substr as overlay,
    split_with_split as split,
    regexp_extract_only_one_group as regexp_extract,
    hex_casted_as_bytes as hex,
    bit_length_from_length as bit_length,
    element_at_using_brackets as element_at,
    array_union_using_array_concat as array_union,
    sequence_from_generate_array as sequence,
    position_as_strpos as position,
    try_to_timestamp_safe as try_to_timestamp,
    _is_string_using_typeof_string as _is_string,
    array_append_using_array_cat as array_append,
    endswith_with_underscore as endswith,
)


def typeof(col: ColumnOrName) -> Column:
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.typeof", expressions=[Column.ensure_col(col).expression]
        )
    )


def degrees(col: ColumnOrName) -> Column:
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.degrees", expressions=[Column.ensure_col(col).expression]
        )
    )


def radians(col: ColumnOrName) -> Column:
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.radians", expressions=[Column.ensure_col(col).expression]
        )
    )


def bround(col: ColumnOrName, scale: t.Optional[int] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    lit = get_func_from_session("lit", _BaseSession())

    expressions = [Column.ensure_col(col).cast("bignumeric").expression]
    if scale is not None:
        expressions.append(lit(scale).expression)
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_round_half_even",
            expressions=expressions,
        )
    )


def months_between(
    date1: ColumnOrName, date2: ColumnOrName, roundOff: t.Optional[bool] = None
) -> Column:
    roundOff = True if roundOff is None else roundOff
    round = get_func_from_session("round")
    lit = get_func_from_session("lit")

    value = Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_months_between",
            expressions=[
                Column.ensure_col(date1).cast("datetime").expression,
                Column.ensure_col(date2).cast("datetime").expression,
            ],
        )
    )
    if roundOff:
        value = round(value, lit(8))
    return value


def next_day(col: ColumnOrName, dayOfWeek: str) -> Column:
    lit = get_func_from_session("lit")

    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_next_day",
            expressions=[Column.ensure_col(col).cast("date").expression, lit(dayOfWeek).expression],
        )
    )


def from_unixtime(col: ColumnOrName, format: t.Optional[str] = None) -> Column:
    from sqlframe.base.session import _BaseSession

    session: _BaseSession = _BaseSession()

    expressions = [Column.ensure_col(col).expression]
    return Column(
        sqlglot_expression.Anonymous(
            this="FORMAT_TIMESTAMP",
            expressions=[
                session.format_time(format),
                Column(
                    sqlglot_expression.Anonymous(this="TIMESTAMP_SECONDS", expressions=expressions)
                ).expression,
            ],
        )
    )


def unix_timestamp(
    timestamp: t.Optional[ColumnOrName] = None, format: t.Optional[str] = None
) -> Column:
    from sqlframe.base.session import _BaseSession

    lit = get_func_from_session("lit")
    return Column(
        sqlglot_expression.Anonymous(
            this="UNIX_SECONDS",
            expressions=[
                sqlglot_expression.Anonymous(
                    this="PARSE_TIMESTAMP",
                    expressions=[
                        _BaseSession().format_time(format),
                        Column.ensure_col(timestamp).expression,
                        lit("UTC").expression,
                    ],
                )
            ],
        )
    )


def format_number(col: ColumnOrName, d: int) -> Column:
    round = get_func_from_session("round")
    lit = get_func_from_session("lit")

    return Column(
        sqlglot_expression.Anonymous(
            this="FORMAT",
            expressions=[
                lit(f"%'.{d}f").expression,
                round(Column.ensure_col(col).cast("float"), d).expression,
            ],
        )
    )


def substring_index(str: ColumnOrName, delim: str, count: int) -> Column:
    lit = get_func_from_session("lit")

    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_substring_index",
            expressions=[
                Column.ensure_col(str).expression,
                lit(delim).expression,
                lit(count).expression,
            ],
        )
    )


def bin(col: ColumnOrName) -> Column:
    return (
        Column(
            sqlglot_expression.Anonymous(
                this="bqutil.fn.to_binary",
                expressions=[Column.ensure_col(col).expression],
            )
        )
        .cast("int")
        .cast("string")
    )


def slice(
    x: ColumnOrName, start: t.Union[ColumnOrName, int], length: t.Union[ColumnOrName, int]
) -> Column:
    lit = get_func_from_session("lit")

    start_col = start if isinstance(start, Column) else lit(start)
    length_col = length if isinstance(length, Column) else lit(length)

    subquery = (
        sqlglot_expression.select(
            sqlglot_expression.column("x"),
        )
        .from_(
            sqlglot_expression.Unnest(
                expressions=[Column.ensure_col(x).expression],
                alias=sqlglot_expression.TableAlias(
                    columns=[sqlglot_expression.to_identifier("x")],
                ),
                offset=sqlglot_expression.to_identifier("offset"),
            )
        )
        .where(
            sqlglot_expression.Between(
                this=sqlglot_expression.column("offset"),
                low=(start_col - lit(1)).expression,
                high=(start_col + length_col).expression,
            )
        )
    )

    return Column(
        sqlglot_expression.Anonymous(
            this="ARRAY",
            expressions=[subquery],
        )
    )


def array_position(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    lit = get_func_from_session("lit")

    value_col = value if isinstance(value, Column) else lit(value)

    return Column(
        sqlglot_expression.Coalesce(
            this=sqlglot_expression.Anonymous(
                this="bqutil.fn.find_in_set",
                expressions=[
                    value_col.expression,
                    sqlglot_expression.Anonymous(
                        this="ARRAY_TO_STRING",
                        expressions=[Column.ensure_col(col).expression, lit(",").expression],
                    ),
                ],
            ),
            expressions=[lit(0).expression],
        )
    )


def array_remove(col: ColumnOrName, value: ColumnOrLiteral) -> Column:
    lit = get_func_from_session("lit")

    value_col = value if isinstance(value, Column) else lit(value)

    filter_subquery = sqlglot_expression.select(
        "*",
    ).from_(
        sqlglot_expression.Unnest(
            expressions=[Column.ensure_col(col).expression],
            alias=sqlglot_expression.TableAlias(
                columns=[sqlglot_expression.to_identifier("x")],
            ),
        )
    )

    agg_subquery = (
        sqlglot_expression.select(
            sqlglot_expression.Anonymous(
                this="ARRAY_AGG",
                expressions=[sqlglot_expression.column("x")],
            ),
        )
        .from_(filter_subquery.subquery("t"))
        .where(
            sqlglot_expression.NEQ(
                this=sqlglot_expression.column("x", "t"),
                expression=value_col.expression,
            )
        )
    )

    return Column(agg_subquery.subquery())


def array_distinct(col: ColumnOrName) -> Column:
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_array_distinct",
            expressions=[Column.ensure_col(col).expression],
        )
    )


def array_min(col: ColumnOrName) -> Column:
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_array_min",
            expressions=[Column.ensure_col(col).expression],
        )
    )


def array_max(col: ColumnOrName) -> Column:
    return Column(
        sqlglot_expression.Anonymous(
            this="bqutil.fn.cw_array_max",
            expressions=[Column.ensure_col(col).expression],
        )
    )


def sort_array(col: ColumnOrName, asc: t.Optional[bool] = None) -> Column:
    order = "ASC" if asc or asc is None else "DESC"
    subquery = (
        sqlglot_expression.select("x")
        .from_(
            sqlglot_expression.Unnest(
                expressions=[Column.ensure_col(col).expression],
                alias=sqlglot_expression.TableAlias(
                    columns=[sqlglot_expression.to_identifier("x")],
                ),
            )
        )
        .order_by(f"x {order}")
    )

    return Column(sqlglot_expression.Anonymous(this="ARRAY", expressions=[subquery]))


array_sort = sort_array
